use std::{
    borrow::Cow,
    cmp,
    collections::{BTreeSet, HashMap, HashSet},
};

use indexmap::IndexMap;
use petgraph::{
    algo::toposort,
    graph::DiGraph,
    prelude::{EdgeIndex, NodeIndex},
};
use quick_xml::{events::attributes::Attribute, name::QName, ElementWriter};

use crate::{
    types::{
        graph::{edge::Edge, FilterGraph},
        nodes::{primitives::WriteElement, CommonAttrs},
    },
    Node,
};

use self::error::CodegenError;

pub struct SvgDocument {
    filters: HashMap<String, FilterGraph>,
}

impl SvgDocument {
    pub fn new() -> Self {
        Self {
            filters: HashMap::new(),
        }
    }

    #[allow(clippy::unwrap_used, reason = "we literally just did the insertion")]
    pub fn create_filter(&mut self, id: impl ToString) -> &mut FilterGraph {
        let filter = FilterGraph::new();

        self.filters.insert(id.to_string(), filter);
        self.filters.get_mut(&id.to_string()).unwrap()
    }

    pub fn generate_svg(&self) -> String {
        let mut result = Vec::new();
        let mut doc_writer = quick_xml::Writer::new_with_indent(&mut result, b' ', 4);

        doc_writer
            .create_element("svg")
            .write_inner_content(|writer| {
                self.filters
                    .iter()
                    .try_fold(writer, Self::gen_filter)
                    .map(|_| {})
            });

        String::from_utf8_lossy(&result).to_string()
    }

    fn gen_filter<'w, 'r>(
        writer: &'w mut quick_xml::Writer<&'r mut Vec<u8>>,
        (id, graph): (&String, &FilterGraph),
    ) -> Result<&'w mut quick_xml::Writer<&'r mut Vec<u8>>, CodegenError> {
        writer
            .create_element("filter")
            .with_attribute(("id", id.as_str()))
            .write_inner_content(|writer| Self::graph_to_svg(writer, graph))
    }

    fn graph_to_svg(
        writer: &mut quick_xml::Writer<&mut Vec<u8>>,
        graph: &FilterGraph,
    ) -> Result<(), CodegenError> {
        let sorted = toposort(&graph.dag, None).expect("no cycles allowed in a DAG");
        sorted
            .into_iter()
            .filter_map(|node_idx| {
                graph
                    .dag
                    .node_weight(node_idx)
                    .and_then(|node| node.primitive())
                    .map(|(primitive, common_attrs)| (node_idx, primitive, common_attrs))
            })
            .try_fold(writer, |writer, (node_idx, primitive, common_attrs)| {
                let mut el_writer = primitive.element_writer(writer);
                el_writer = input_attrs(node_idx, &graph.dag).write_into(el_writer);
                el_writer = output_attrs(node_idx, &graph.dag).write_into(el_writer);
                el_writer = Attrs::from(*common_attrs).write_into(el_writer);
                el_writer.write_empty()
            })?;

        Ok(())
    }
}

struct Attrs(IndexMap<String, String>);

impl Attrs {
    fn write_into<'w, 'b>(
        self,
        el_writer: ElementWriter<'w, &'b mut Vec<u8>>,
    ) -> ElementWriter<'w, &'b mut Vec<u8>> {
        let attrs = self.0.iter().map(|(k, v)| (k.as_str(), v.as_str()));
        el_writer.with_attributes(attrs)
    }
}

impl From<CommonAttrs> for Attrs {
    fn from(value: CommonAttrs) -> Self {
        let CommonAttrs {
            x,
            y,
            width,
            height,
        } = value;
        Self(IndexMap::from([
            ("x".to_owned(), x.to_string()),
            ("y".to_owned(), y.to_string()),
            ("width".to_owned(), width.to_string()),
            ("height".to_owned(), height.to_string()),
        ]))
    }
}

#[allow(
    clippy::unwrap_used,
    reason = "in all cases of use of unwrap, it's values we got from the graph so we know it's safe"
)]
fn input_attrs(node_idx: NodeIndex, g: &DiGraph<Node, Edge>) -> Attrs {
    let inputs: Vec<NodeIndex> = g
        .neighbors_directed(node_idx, petgraph::Direction::Incoming)
        .collect();
    let node = g.node_weight(node_idx).unwrap();

    if node.input_count() as usize != inputs.len() {
        todo!("proper error handling for wrong numbers of inputs")
    }

    let mut inputs = inputs
        .into_iter()
        .map(|input_idx| {
            let edge_idx = g.find_edge(input_idx, node_idx).unwrap();
            let edge = g.edge_weight(edge_idx).unwrap();
            let input_node = g.node_weight(input_idx).unwrap();

            (input_node, edge_idx, edge)
        })
        .collect::<Vec<(&Node, EdgeIndex, &Edge)>>();
    inputs.sort_by(|a, b| a.2.cmp(b.2));

    let mut uniq = BTreeSet::new();
    let no_duplicates = inputs.iter().all(|(_, _, edge)| uniq.insert(edge));

    if no_duplicates {
        Attrs(
            inputs
                .into_iter()
                .map(|(input, edge_idx, edge)| {
                    let name = match input {
                        Node::StdInput(s) => format!("{s:?}"),
                        Node::Primitive { .. } => format_edge_idx(edge_idx),
                    };
                    (edge.to_string(), name)
                })
                .collect(),
        )
    } else {
        todo!("better error handling for inputs with multiple values")
    }
}

#[allow(
    clippy::unwrap_used,
    reason = "in all cases of use of unwrap, it's values we got from the graph so we know it's safe"
)]
fn output_attrs(node_idx: NodeIndex, g: &DiGraph<Node, Edge>) -> Attrs {
    let outputs: Vec<NodeIndex> = g
        .neighbors_directed(node_idx, petgraph::Direction::Outgoing)
        .collect();

    match outputs.len().cmp(&1) {
        cmp::Ordering::Less => Attrs(IndexMap::new()),
        cmp::Ordering::Equal => {
            let output = outputs.first().unwrap();
            let edge_idx = g.find_edge(node_idx, *output).unwrap();
            Attrs(IndexMap::from([(
                "result".to_string(),
                format_edge_idx(edge_idx),
            )]))
        }
        cmp::Ordering::Greater => todo!("better error handling for too many outputs"),
    }
}

/// convenience method to avoid fuckups during future changes
fn format_edge_idx(idx: EdgeIndex) -> String {
    format!("edge{}", idx.index())
}

mod error {
    use std::{error::Error, fmt::Display};

    #[derive(Debug)]
    pub enum CodegenError {
        QuickXmlError(quick_xml::Error),
    }

    impl From<quick_xml::Error> for CodegenError {
        fn from(value: quick_xml::Error) -> Self {
            Self::QuickXmlError(value)
        }
    }

    impl Display for CodegenError {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                CodegenError::QuickXmlError(e) => e.fmt(f),
            }
        }
    }

    impl Error for CodegenError {}
}

impl Default for SvgDocument {
    fn default() -> Self {
        Self::new()
    }
}
